/*	$OpenBSD: morse.c,v 1.22 2016/03/07 12:07:56 mestre Exp $	*/

/*
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <ctype.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "cw.h"

static char
	*digit[] = {
	"-----",
	".----",
	"..---",
	"...--",
	"....-",
	".....",
	"-....",
	"--...",
	"---..",
	"----.",
},
	*alph[] = {
	".-",
	"-...",
	"-.-.",
	"-..",
	".",
	"..-.",
	"--.",
	"....",
	"..",
	".---",
	"-.-",
	".-..",
	"--",
	"-.",
	"---",
	".--.",
	"--.-",
	".-.",
	"...",
	"-",
	"..-",
	"...-",
	".--",
	"-..-",
	"-.--",
	"--..",
};

struct punc {
	char c;
	char *morse;
} other[] = {
	{ 'e', "..-.." },	/* accented e - only decodes */
	{ ',', "--..--" },
	{ '.', ".-.-.-" },
	{ '?', "..--.." },
	{ '/', "-..-." },
	{ '-', "-....-" },
	{ ':', "---..." },
	{ ';', "-.-.-." },
	{ '(', "-.--." },	/* KN */
	{ ')', "-.--.-" },
	{ '"', ".-..-." },
	{ '`', ".-..-." },
	{ '\'', ".----." },
	{ '+', ".-.-." },	/* AR \n\n\n */
	{ '=', "-...-" },	/* BT \n\n */
	{ '@', ".--.-." },
	{ '&', ".-..." },	/* AS wait - ampersand for encoding only */
	{ '\n', ".-.-" },	/* AA (will only decode) */
	{ '\a', "-.-.-" },	/* KA attention - BEL for encoding only */
	{ '\b', "........" },	/* HH correction - BS for encoding only */
	{ '\6', "...-." },	/* SN understood - ACK for encoding only */
	{ '\0', NULL }
};

/* this list takes precedence over 'other' for decoding */
struct prosign {
	char *c;
	char *morse;
} ps[] = {
	{ "<AS>", ".-..." },	/* wait (unofficial: ampersand) */
	{ "<CL>", "-.-..-.." },	/* closing station (unofficial) */
	{ "<E5>", "......" },	/* error (unofficial) */
	{ "<EE5>", "......." },	/* error (unofficial) */
	{ "<HH>", "........" },	/* correction */
	{ "<KA>", "-.-.-" },	/* attn. / start of transmission, aka. <CT> */
	{ "<SK>", "...-.-" },	/* end of transmission */
	{ "<SN>", "...-." },	/* understood, aka. <VE> */
	{ "<SOS>", "...---..." },
	{ NULL, NULL }
};

void	morse(int);
void	parse(int);
int	lookup(char *);
void	decode(char *);
void	show(char *);
void	sound(char *);

static int sflag = 0;
static int dflag = 0;
static int aflag = 0;

/* audio parameters */
static int speed = 24;		/* base speed [wpm] */
static int pspeed = 24;		/* speed used for pauses [wpm] */
static int freq = 600;		/* tone pitch [Hz] */
static int ditlen, pditlen;	/* dit length, separately for pauses [ms] */

int
main(int argc, char *argv[])
{
	int ch;
	char *p;

	/* pledge set needed for stdio and sndio */
	if (pledge("stdio rpath wpath cpath inet unix dns audio", NULL) == -1)
		err(1, "pledge");

	while ((ch = getopt(argc, argv, "adsh")) != -1)
		switch(ch) {
		case 'a':
			aflag = 1;
			break;
		case 'd':
			dflag = 1;
			break;
		case 's':
			sflag = 1;
			break;
		case 'h':
		default:
			fprintf(stderr,
			    "usage: %s [-a] [-d | -s] [string ...]\n",
			    getprogname());
			return 1;
		}
	argc -= optind;
	argv += optind;

	/* audio initialization */
	if (aflag) {
		/* environment variables */
		getenvnum("MORSE_SPEED", "wpm", &speed);
		pspeed = speed;	/* default */
		getenvnum("MORSE_PAUSE_SPEED", "wpm", &pspeed);
		getenvnum("MORSE_PITCH", "Hz", &freq);

		/* derive base (dit) lengths from speed settings */
		ditlen = 1200 / speed;
		pditlen = 1200 / pspeed;

		cwinit();

		/* after sio_open() only "audio" remains needed */
		if (pledge("stdio audio", NULL) == -1)
			err(1, "pledge");

		cwstart();
	} else {
		/* no audio - drop access to all but stdio */
		if (pledge("stdio", NULL) == -1)
			err(1, "pledge");
	}

	setvbuf(stdout, NULL, _IOLBF, 0);	/* always use line buffering */
	if (dflag) {
		if (*argv) {
			decode(*argv);
			while (*++argv) {
				decode("");	/* inter-word space */
				decode(*argv);
			}
		} else while ((ch = getchar()) != EOF)
			parse(ch);
		putchar('\n');
	} else {
		if (*argv)
			do {
				for (p = *argv; *p; ++p)
					morse((int)*p);
				show("");
			} while (*++argv);
		else while ((ch = getchar()) != EOF)
			morse(ch);
		show("...-.-");	/* SK */
	}
	if (aflag)
		cwstop();
	return 0;
}

void
morse(int c)
{
	int i;

	if (isalpha(c))
		show(alph[c - (isupper(c) ? 'A' : 'a')]);
	else if (isdigit(c))
		show(digit[c - '0']);
	else if (isspace(c))
		show("");  /* could show BT for a pause */
	else {
		i = 0;
		while (other[i].c) {
			if (other[i].c == c) {
				show(other[i].morse);
				break;
			}
			i++;
		}
	}
}

void
parse(int ch)
{
	static char foo[10];	/* All morse chars shorter than this */
	static int isblank, i;

	if (ch == '-' || ch == '.') {
		foo[i++] = ch;
		if (i == 10) {
			/* overrun means gibberish--print 'x' and advance */
			i = 0;
			putchar('x');
			while ((ch = getchar()) != EOF &&
			    (ch == '.' || ch == '-'))
				;
			isblank = 1;
		}
	} else if (i) {
		foo[i] = '\0';
		decode(foo);
		i = 0;
		isblank = 0;
	} else if (isspace(ch)) {
		if (isblank) {
			/* print whitespace for each double blank */
			decode("");
			isblank = 0;
		} else
			isblank = 1;
	}
}



void
decode(char *s)
{
	int x;

	if (*s)
		x = lookup(s);
	else
		x = ' ';			/* s = "" means inter-word */

	if (x == 0) {
		putchar('x');			/* line noise */
		return;
	}

	if (aflag)
		sound(s);			/* play back iff decoded */

	if (x < 0)				/* negative means prosign */
		printf(" %s ", ps[-x].c);	/* surround with whitespace */
	else
		putchar(x);
}

int
lookup(char *s)
{
	int i;

	for (i = 0; i < 10; i++)
		if (strcmp(digit[i], s) == 0)
			return '0' + i;

	for (i = 0; i < 26; i++)
		if (strcmp(alph[i], s) == 0)
			return 'A' + i;

	for (i = 0; ps[i].c; i++)
		if (strcmp(ps[i].morse, s) == 0)
			return -i;		/* negative means prosign */

	for (i = 0; other[i].c; i++)
		if (strcmp(other[i].morse, s) == 0)
			return other[i].c;

	return 0;
}

void
show(char *s)
{
	char *p;
	int daw = 0;

	putchar(' ');
	if (sflag)
		fputs(s, stdout);
	else if (*s) {
		for (p = s; *p; ++p) {
			if (daw)
				putchar('-');
			if (*p == '.') {
				fputs("di", stdout);
				daw = 0;
			} else if (*p == '-') {
				fputs("daw", stdout);
				daw = 1;
			}
		}
		if (!daw)
			putchar('t');
	}
	if (aflag)
		sound(s);
	printf("\n");
}

void
sound(char *s)
{
	if (*s == '\0')
		silence(7 * pditlen - ditlen);	/* inter-word pause */
	else
		silence(3 * pditlen - ditlen);	/* intra-word pause */

	while (*s) {
		if (*s == '.')
			beep(freq, ditlen);
		else if (*s == '-')
			beep(freq, 3 * ditlen);
		s++;
		silence(ditlen);
	}
}
